import torch
import torch.nn.functional as f
from onpolicy.algorithms.r_mappo.algorithm.r_actor_critic import R_Actor, R_Critic
from onpolicy.utils.util import update_linear_schedule, soft_update


class R_MAPPOPolicy:
    """
    MAPPO Policy  class. Wraps actor and critic networks to compute actions and value function predictions.

    :param args: (argparse.Namespace) arguments containing relevant model and policy information.
    :param obs_space: (gym.Space) observation space.
    :param cent_obs_space: (gym.Space) value function input space (centralized input for MAPPO, decentralized for IPPO).
    :param action_space: (gym.Space) action space.
    :param device: (torch.device) specifies the device to run on (cpu/gpu).
    """

    def __init__(self, args, obs_space, cent_obs_space, act_space, device=torch.device("cpu")):
        self.device = device
        self.lr = args.lr
        self.critic_lr = args.critic_lr
        self.opti_eps = args.opti_eps
        self.weight_decay = args.weight_decay
        self.args = args

        self.obs_space = obs_space
        self.share_obs_space = cent_obs_space
        self.act_space = act_space

        self.actor = R_Actor(args, self.obs_space, self.act_space, self.device)
        if args.if_use_independent_ppo:
            self.critic = R_Critic(args, obs_space, device)
        elif args.if_use_CoPPO and args.use_adv:
            self.critic = R_Critic(args, self.share_obs_space, self.device)
            self.trgt_critic = R_Critic(args, self.share_obs_space, self.device)
            self.trgt_critic.load_state_dict(self.critic.state_dict())
        else:
            self.critic = R_Critic(args, self.share_obs_space, self.device)

        self.actor_optimizer = torch.optim.Adam(self.actor.parameters(),
                                                lr=self.lr, eps=self.opti_eps,
                                                weight_decay=self.weight_decay)
        self.critic_optimizer = torch.optim.Adam(self.critic.parameters(),
                                                 lr=self.critic_lr,
                                                 eps=self.opti_eps,
                                                 weight_decay=self.weight_decay)

    def lr_decay(self, episode, episodes):
        """
        Decay the actor and critic learning rates.
        :param episode: (int) current training episode.
        :param episodes: (int) total number of training episodes.
        """
        update_linear_schedule(self.actor_optimizer, episode, episodes, self.lr)
        update_linear_schedule(self.critic_optimizer, episode, episodes, self.critic_lr)

    def get_actions(self, cent_obs, obs, rnn_states_actor, rnn_states_critic, masks, available_actions=None,
                    deterministic=False, trgt_rnn_states_critic=None):
        """
        Compute actions and value function predictions for the given inputs.
        :param cent_obs (np.ndarray): centralized input to the critic.
        :param obs (np.ndarray): local agent inputs to the actor.
        :param rnn_states_actor: (np.ndarray) if actor is RNN, RNN states for actor.
        :param rnn_states_critic: (np.ndarray) if critic is RNN, RNN states for critic.
        :param masks: (np.ndarray) denotes points at which RNN states should be reset.
        :param available_actions: (np.ndarray) denotes which actions are available to agent
                                  (if None, all actions available)
        :param deterministic: (bool) whether the action should be mode of distribution or should be sampled.

        :return values: (torch.Tensor) value function predictions.
        :return actions: (torch.Tensor) actions to take.
        :return action_log_probs: (torch.Tensor) log probabilities of chosen actions.
        :return rnn_states_actor: (torch.Tensor) updated actor network RNN states.
        :return rnn_states_critic: (torch.Tensor) updated critic network RNN states.
        """
        if self.args.if_use_CoPPO and self.args.use_adv:
            actions, action_log_probs, rnn_states_actor, pi_dist = self.actor(obs,
                                                                 rnn_states_actor,
                                                                 masks,
                                                                 available_actions,
                                                                 deterministic)
        else:
            actions, action_log_probs, rnn_states_actor = self.actor(obs,
                                                                    rnn_states_actor,
                                                                    masks,
                                                                    available_actions,
                                                                    deterministic)
        if self.args.if_use_independent_ppo:
            values, rnn_states_critic = self.critic(obs, rnn_states_critic, masks)
        elif self.args.if_use_CoPPO and self.args.use_adv:
            if self.args.use_trgt:
                q_next_target, trgt_rnn_states_critic = self.trgt_critic((cent_obs, actions), trgt_rnn_states_critic, masks)
            else:
                q_next_target, _ = self.critic((cent_obs, actions), rnn_states_critic, masks)
            values, rnn_states_critic = self.critic((cent_obs, actions), rnn_states_critic, masks)
            if not self.args.use_trgt:
                trgt_rnn_states_critic = rnn_states_critic
            return values, q_next_target, actions, action_log_probs, rnn_states_actor, rnn_states_critic, pi_dist, trgt_rnn_states_critic
        else:                
            values, rnn_states_critic = self.critic(cent_obs, rnn_states_critic, masks)            
        return values, actions, action_log_probs, rnn_states_actor, rnn_states_critic

    def get_values(self, cent_obs, rnn_states_critic, masks):
        """
        Get value function predictions.
        :param cent_obs (np.ndarray): centralized input to the critic.
        :param rnn_states_critic: (np.ndarray) if critic is RNN, RNN states for critic.
        :param masks: (np.ndarray) denotes points at which RNN states should be reset.

        :return values: (torch.Tensor) value function predictions.
        """
        values, _ = self.critic(cent_obs, rnn_states_critic, masks)
        return values

    def get_q_values(self, cent_obs, actions, rnn_states_critic, masks):
        inputs = (cent_obs, actions)
        q_values, _ = self.critic(inputs, rnn_states_critic, masks)
        return q_values

    def get_trgt_q_values(self, cent_obs, actions, masks, trgt_rnn_states_critic=None):
        inputs = (cent_obs, actions)
        trgt_q_values, _ = self.trgt_critic(inputs, trgt_rnn_states_critic, masks)
        return trgt_q_values

    def evaluate_actions(self, cent_obs, obs, rnn_states_actor, rnn_states_critic, action, masks,
                         available_actions=None, active_masks=None, trgt_rnn_states_critic=None):
        """
        Get action logprobs / entropy and value function predictions for actor update.
        :param cent_obs (np.ndarray): centralized input to the critic.
        :param obs (np.ndarray): local agent inputs to the actor.
        :param rnn_states_actor: (np.ndarray) if actor is RNN, RNN states for actor.
        :param rnn_states_critic: (np.ndarray) if critic is RNN, RNN states for critic.
        :param action: (np.ndarray) actions whose log probabilites and entropy to compute.
        :param masks: (np.ndarray) denotes points at which RNN states should be reset.
        :param available_actions: (np.ndarray) denotes which actions are available to agent
                                  (if None, all actions available)
        :param active_masks: (torch.Tensor) denotes whether an agent is active or dead.

        :return values: (torch.Tensor) value function predictions.
        :return action_log_probs: (torch.Tensor) log probabilities of the input actions.
        :return dist_entropy: (torch.Tensor) action distribution entropy for the given inputs.
        """
        action_log_probs, dist_entropy = self.actor.evaluate_actions(obs,
                                                                    rnn_states_actor,
                                                                    action,
                                                                    masks,
                                                                    available_actions,
                                                                    active_masks)
        if self.args.if_use_independent_ppo:
            values, _ = self.critic(obs, rnn_states_critic, masks)
            return values, action_log_probs, dist_entropy
        elif self.args.if_use_CoPPO and self.args.use_adv:
            values, _ = self.critic((cent_obs, action), rnn_states_critic, masks) # (n_r_t*ep_len*n_ags, 1)
            q_next_target, _ = self.trgt_critic((cent_obs, action), trgt_rnn_states_critic, masks)
            return values, q_next_target, action_log_probs, dist_entropy
        else:
            values, _ = self.critic(cent_obs, rnn_states_critic, masks)
            return values, action_log_probs, dist_entropy

    def act(self, obs, rnn_states_actor, masks, available_actions=None, deterministic=False):
        """
        Compute actions using the given inputs.
        :param obs (np.ndarray): local agent inputs to the actor.
        :param rnn_states_actor: (np.ndarray) if actor is RNN, RNN states for actor.
        :param masks: (np.ndarray) denotes points at which RNN states should be reset.
        :param available_actions: (np.ndarray) denotes which actions are available to agent
                                  (if None, all actions available)
        :param deterministic: (bool) whether the action should be mode of distribution or should be sampled.
        """
        if self.args.if_use_CoPPO and self.args.use_adv:
            actions, _, rnn_states_actor, pi_dist = self.actor(obs, rnn_states_actor, masks, available_actions, deterministic)
            return actions, rnn_states_actor, pi_dist
        else:
            actions, _, rnn_states_actor = self.actor(obs, rnn_states_actor, masks, available_actions, deterministic)
            return actions, rnn_states_actor
